/* 
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:   etftp.h
 * Module:      etftp
 * Created:     0.1, 970421, Torbjrn Frostemark
 * Revised:     1.1, 980219, Torbjrn Frostemark
 * 
 * Description: Declarations needed for ETFTP client and server
 * Remarks:     -
 *----------------------------------------------------------------------
 */

#ifndef _ETFTP
#define _ETFTP

/* #undef LINUX_ETFTP */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
/*#include <sys/uio.h> */
#include <unistd.h>
#include <setjmp.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ctype.h>
#include <sys/unistd.h>

#include "defs.h"		/* contains all #define needed for etftp */

#ifdef CLIENT_OR_SERVER

/*
 * ETFTP client and server stuff
 */

extern void breakpoint();

/* Externals. */
extern char     *command;       /* the command being processed */
extern int      connected;      /* true if we're connected to host */
extern char     *hostName;      /* name of host system */
extern int      inetdFlag;      /* true if we were started by a daemon */
extern int      interactive;    /* true if we're running interactive */
extern jmp_buf  jmpMainLoop;    /* to return to main command loop */
extern int      lastSend;       /* #bytes of data in last data packet */
extern FILE     *localFP;       /* fp of local file to read or write */
extern int      modeType;       /* see MODE_xxx values */
extern int      nextBlkNum;     /* next block# to send/receive */
extern char     *pName;         /* the name by which we are invoked */
extern char     *prompt;        /* prompt string, for interactive use */
extern long     totNBytes;      /* for get/put statistics printing */

/* N E W S  */
extern unsigned short	blockSize;	/* Current data-block size */
extern unsigned short	bufferSize;	/* No of blocks per buffer */
extern unsigned short	burstSize;	/* Current burst size */
extern int	multiBuff;	/* No of buffers */
extern long	baudRate;	/* transfer speed */
extern unsigned short	radioDelay;	/* Estimated turnover time */
extern char	tempToken[];	/* temporary token for anyone to use */

/* M O R E   N E W S */
extern int	firstBlkNum;	/* First packet in buffer */
extern long	bufferNum;	/* Current buffer number */
extern long	connectID;	/* Unique connection ID */
extern short	renegotiate;
extern short	localPort;
extern short	remotePort;
extern int      outSD;		/* Out socket */

#ifdef CLIENT
extern short	sesRemotePort;  /* Remote port for current session */
extern int	hashFlag;	/* -h command line option, or "hash" cmd */
extern int	ctrlTime;	/* when Ctrl-C is pressed */
extern long	bytesReceived;	/* bytes received counter */
#endif /* CLIENT */

#ifdef SERVER
extern int	transferMode;	/* put/get. Used to measure server rtt */
extern int	rttSet;		/* Used to measure server rtt */
#endif /* SERVER */

extern unsigned short	burstRate;
extern short	sequenceNum;
extern short	tightTimer;
extern short	looseTimer;
extern long	rtTimer;	/* round trip timer */
extern int	recvnBytes;	/* No of bytes recieved by NetListen */
extern unsigned short packetSCounter; /* counts number of packets sent */
extern unsigned short packetRCounter; /* counts number of packets received */

/*
 * One receive buffer and one transmit buffer.
 */

extern char     recvBuff[MAXBUFSZ*MAXBLOCKSZ];
extern char     sendBuff[MAXBUFSZ*MAXBLOCKSZ];
extern char	*sendPack;
extern char	*recvPack;
extern short      sendLen;      /* number of bytes in sendPack[] */

extern int      pkSent;         /* last packettype sent */
extern int      pkRecv;         /* last packettype received */

/*
 * Datatypes of functions that don't return an int.
 */

extern char     *GetToken();
extern double   t_GetRTime();   /* our library routine to return elapsed time */
extern char     *SysErrStr();   /* our library routine for system error messages */


/*
 * Abstract data types.
 */

typedef struct RecvStat {       /* client and server statistics */
  double        aveRate;        /* average data rate */
  double        momRate;        /* momentary data rate */
  double        aveDel;         /* average delay */
  double        momDel;         /* momentary delay */
  double        aveLoss;        /* average packet loss */
  double        momLoss;        /* momentary packet loss */
} RecvStat;

/*
 * A structure used by TRON.
 */

typedef struct TRONList {
    short *list;
    int size;
 } TRONList;

#endif /* CLIENT_OR_SERVER */


#ifdef SERVER
/*
 * Server stuff.
 */

/* externals */
extern int timeoutFlag;		/* net timeout */
#endif /* SERVER */

/*
 * Declarations used in all three components
 */

/* Externals. */
extern int      traceFlag;      /* -t command line option, or "trace" cmd */
extern int	verboseFlag;	/* -v command line option, or "verbose" cmd */
#ifndef LINUX_ETFTP
typedef void(*sighandler_t)(int); /* Pure magic. Don't ask. */
#endif /* LINUX_ETFTP */
/* 
/usr/include/stdlib.h:#define   EXIT_FAILURE    1
/usr/include/stdlib.h:#define   EXIT_SUCCESS    0
*/

#endif /* _ETFTP */





