/* 
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:   fsm.c
 * Module:      FSM (Finite State Machine)
 * Created:     0.1, 970503, Henrik Brodin
 * Revised:     1.1, 980219, Torbjrn Frostemark
 * 
 * Description: Finite state machine routines.
 * Remarks:     -
 *
 * Private Methods: 	FSMInvalid()
 *			FuncTimeOut()
 *			FSMName(int num)
 *
 * Private Attributes: 	rTTInfo
 *			rTTFirst
 *----------------------------------------------------------------------
 */

#include        "clu.h"
#include	"fsm.h"
#include	"rtt.h"		/* for RTT timing */
#include	"net.h"
#include	"sendrecv.h"

/*
 * Private attributes.
 */

static	struct RTTStruct	rTTInfo; /* used by the RTT functions */
static	int		   	rTTFirst = 1;

/*
 * Finite state machine table.
 * This is just a 2-d array indexed by the last opcode sent and
 * the opcode just received.  The result is the address of a
 * function to call to process the received opcode.
 */

typedef int (*function)(); 

#ifdef CLIENT
function fSMPtr [ PT_MAX + 1 ] [ PT_MAX + 1 ] = {
  {
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_OPEN]		*/
    RecvRESPONSE,	/* [sent = PT_OPEN]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_OPEN]  [recv = PT_CONTROL]	*/
    RecvREFUSED,	/* [sent = PT_OPEN]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_OPEN]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_RESPONSE]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_RESPONSE]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_KEEPALIVE]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_KEEPALIVE]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_QUIT]		*/
    RecvQUITACK,	/* [sent = PT_QUIT]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_QUIT]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_REFUSED]	*/
    RecvDONE		/* [sent = PT_QUIT]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_QUITACK]  [recv = PT_ABORT]     	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_DATA]      	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_LDATA]     	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_QUITACK]  [recv = PT_DONE]      	*/
  }, {
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_ABORT]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_DATA]  [recv = PT_ABORT]		*/
    RecvDATA,		/* [sent = PT_DATA]  [recv = PT_DATA]		*/
    RecvLDATA,		/* [sent = PT_DATA]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_DATA]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_DATA]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_LDATA]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_LDATA]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_LDATA]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_QUITACK]	*/
    RecvABORT,	        /* [sent = PT_NULL_ACK]  [recv = PT_ABORT]	*/
    RecvDATA,		/* [sent = PT_NULL_ACK]  [recv = PT_DATA]	*/
    RecvLDATA,		/* [sent = PT_NULL_ACK]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_NULL_ACK]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_REFUSED]	*/
    RecvDONE		/* [sent = PT_NULL_ACK]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_KEEPALIVE]	*/
    RecvQUIT,		/* [sent = PT_CONTROL]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_CONTROL]  [recv = PT_ABORT]	*/
    RecvDATA,		/* [sent = PT_CONTROL]  [recv = PT_DATA]	*/
    RecvLDATA,		/* [sent = PT_CONTROL]  [recv = PT_LDATA]	*/
    RecvNULL_ACK,	/* [sent = PT_CONTROL]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_CONTROL]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_REFUSED]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_DONE]  [recv = PT_DONE]		*/
  }
};
#endif	/* CLIENT */

#ifdef	SERVER
function	fSMPtr [ PT_MAX + 1 ] [ PT_MAX + 1 ] = {
  {
	RecvOPEN,	/* [sent = PT_OPEN]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_OPEN]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_OPEN]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_RESPONSE]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_RESPONSE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_RESPONSE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_RESPONSE]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_KEEPALIVE]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_KEEPALIVE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_KEEPALIVE]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_QUIT]		*/
    RecvQUITACK,	/* [sent = PT_QUIT]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_QUIT]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_QUIT]  [recv = PT_REFUSED]	*/
    RecvDONE		/* [sent = PT_QUIT]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_QUITACK]  [recv = PT_ABORT]     	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_DATA]      	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_LDATA]     	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_QUITACK]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_QUITACK]  [recv = PT_DONE]      	*/
  }, {
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_ABORT]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_ABORT]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_DATA]  [recv = PT_ABORT]		*/
    RecvDATA,		/* [sent = PT_DATA]  [recv = PT_DATA]		*/
    RecvLDATA,		/* [sent = PT_DATA]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_DATA]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_DATA]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_DATA]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_LDATA]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_LDATA]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_LDATA]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_LDATA]  [recv = PT_DONE]		*/
  }, {
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_NULL_ACK]  [recv = PT_ABORT]	*/
    RecvDATA,		/* [sent = PT_NULL_ACK]  [recv = PT_DATA]	*/
    RecvLDATA,		/* [sent = PT_NULL_ACK]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_NULL_ACK]	*/
    RecvCONTROL,	/* [sent = PT_NULL_ACK]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_NULL_ACK]  [recv = PT_REFUSED]	*/
    RecvDONE		/* [sent = PT_NULL_ACK]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_KEEPALIVE]	*/
    RecvQUIT,		/* [sent = PT_CONTROL]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_QUITACK]	*/
    RecvABORT,		/* [sent = PT_CONTROL]  [recv = PT_ABORT]	*/
    RecvDATA,		/* [sent = PT_CONTROL]  [recv = PT_DATA]	*/
    RecvLDATA,		/* [sent = PT_CONTROL]  [recv = PT_LDATA]	*/
    RecvNULL_ACK,	/* [sent = PT_CONTROL]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_CONTROL]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_CONTROL]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_OPEN]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_QUIT]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_ABORT]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_DATA]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_LDATA]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_REFUSED]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_REFUSED]  [recv = PT_DONE]	*/
  }, {
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_OPEN]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_RESPONSE]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_KEEPALIVE]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_QUIT]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_QUITACK]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_ABORT]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_DATA]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_LDATA]		*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_NULL_ACK]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_CONTROL]	*/
	FSMInvalid,	/* [sent = PT_DONE]  [recv = PT_REFUSED]	*/
	FSMInvalid	/* [sent = PT_DONE]  [recv = PT_DONE]		*/
  }
};
#endif	/* SERVER */

/*
 *----------------------------------------------------------------------
 * CheckSum
 *
 *      Calculates the checksum of the received packet. 
 *
 * Results:
 *	Returns EXIT_SUCCESS or EXIT_FAILURE.
 *
 * Side effects:
 *	-
 *----------------------------------------------------------------------
 */

int CheckSum()
{
  int check, i, length;
  
  /*DEBUG1("FSM: Checksum: recvPack = %p", recvPack);*/
  length = PF_LENGTH(recvPack);

  /*DEBUG1("FSM: checksum: PF_LENGTH returned: %d.", length);*/
  if (length > (1900 + 24)) {
    DEBUG1("FSM: CheckSum: Bogus packet size of %d.", length);
    return EXIT_FAILURE;
  }
  for (i = 0, check = 0; i < length; i+= 2) { 
    /* We're counting in words, hence += 2. */
      check = check ^ ldshort(&recvPack[i]);
    }
  if (check == 0xFFFF) {
    return EXIT_SUCCESS;
  } else {
    return EXIT_FAILURE;
  }
}

/*
 *----------------------------------------------------------------------
 * FSMLoop
 *
 *	Main loop of finite state machine.
 *
 * 	For the client, we're called after either an PUT or a GET has been
 * 	sent to the other side.
 *
 * 	For the server, we're called after either an PUT or a GET has been
 * 	received from the other side.  In this case, the argument will be a
 * 	0 (since nothing has been sent) but the state table above handles
 * 	this. 
 *
 * Results:
 *	return EXIT_SUCCESS on normal termination, EXIT_FAILURE on timeout 
 *
 * Side effects:
 *	-
 *----------------------------------------------------------------------
 */

int FSMLoop(opcode)
     int opcode;		/* for client: PT_OPEN */
				/* for server: PT_OPEN */
{
  register int	nBytes;
  register int	packNum = -1;       	/* for calculating RTT */

  DEBUG1("FSMLoop: Start. opcode = %d", opcode);

  if (rTTFirst) {
    DEBUG0("FSMLoop: Initializing new RTT.");
    RTTInit(&rTTInfo);
    rTTFirst = 0;
  }

  if (opcode==-1) {
    pkSent=0;
    goto serverJUMP;
  } else {
    pkSent = opcode;
  }

  RTTNewPack(&rTTInfo, -1);	/* initialize for a new packet */

  while (TRUE) {
    int tmp;

    signal(SIGALRM, (sighandler_t) FuncTimeOut);
    /*DEBUG0("FSMLoop: Resetting timeoutFlag");*/
    timeoutFlag = FALSE;
    recvPack=malloc(MAXBUFF);/*test... was MAXPACKETSZ*/    /* We want it to be null for checksum reasons. */
    /*DEBUG2("FSMLoop: Allocated %d byte at %p", MAXBUFF, recvPack);*/

    tmp = RTTUpdate(&rTTInfo, packNum);
    /*    DEBUG2("FSMLoop: wait = %d. alarmTiemout = %d", rTTInfo.wait, rTTInfo.alarmTimeout);*/
    DEBUG1("FSMLoop: Calling alarm(%d)", tmp);
    alarm(tmp);       /* calc timeout & start timer */

#ifdef CLIENT
    if ((nBytes = ClientNetRecv(recvPack, MAXBUFF)) < 0) 
#endif /* CLIENT */

#ifdef SERVER
    if ((nBytes = ServerNetRecv(recvPack, MAXBUFF)) < 0) 
#endif /* SERVER */
      {
	if (timeoutFlag) {
	  /*
	   * The receive timed out.  See if we've tried
	   * enough, and if so, return to caller.
	   */
	  if (RTTTimeout(&rTTInfo) < 0) {
	    DEBUG0("FSMLoop: Transfer timed out!");
	    return EXIT_FAILURE;
	  }
	  DEBUG1("FSMLoop: Resend number %d", rTTInfo.rtt_nrexmt);
	} else {
	  DEBUG0("FSMLoop: Nothing received.");
	}

	/*
	 * Retransmit the last packet.
	 */
#ifdef CLIENT
	if (((pkRecv != PT_DATA) && (pkRecv != PT_LDATA)) || ((pkRecv == PT_LDATA) && (pkSent == PT_CONTROL))) {
	  DEBUG1("FSMLoop: Client will resend last packet (type %d)", PF_TYPE(sendPack));
#endif /* CLIENT */

#ifdef SERVER
	if ((pkSent == PT_LDATA) || (pkSent == PT_CONTROL)) {
	  DEBUG1("FSMLoop: Server will resend last packet (type %d)", PF_TYPE(sendPack));
#endif /* SERVER */
	  NetSend(sendPack,sendLen);
	} else {
	  DEBUG0("FSMLoop: Last packet will NOT be resent!");
	}
	continue;
      }

    /*
     * Packet has arrived successfully!
     */

    if (CheckSum() == EXIT_FAILURE) {
      DEBUG0("FSM: Checksum failed!");
      continue;			/* If checksum is not correct, throw away the packet. */
    }
    
    alarm(0);			/* stop signal timer */
    timeoutFlag = FALSE;
    
    if (nBytes < 4) {
      DEBUG1("receive length = %d bytes", nBytes);
    }

  serverJUMP:
    pkRecv = PF_TYPE(recvPack);
    if (pkRecv < PT_MIN || pkRecv > PT_MAX) {
      DEBUG1("Invalid packet type received: %d", pkRecv);
    }
    /*
     * We call the appropriate function, passing the address
     * of the receive buffer and its length.  These arguments
     * ignore the received-opcode, which we've already processed.
     *
     * We assume the called function will send a response to the
     * other side.  It is the called function's responsibility to
     * set pkSent to the op-code that it sends to the other side.
     */
    /*DEBUG0("FSMLoop: Doing FSM jump! (dangerous)");*/
    if ((*fSMPtr[pkSent][pkRecv])() == EXIT_FAILURE) {
      /*
       * When the called function returns EXIT_FAILURE, this loop
       * is done.  Turn off the signal handler for  timeouts 
       * and return to the caller.
       */
      signal(SIGALRM, SIG_DFL);
      if ((pkSent == PT_DONE) || (pkRecv == PT_DONE)) {
	/* Transfer is a cuccess! */
	DEBUG0("FSMLoop: exit success in fsmloop!");
	return EXIT_SUCCESS;
      } else {
	/* Something went wrong during transfer */
	DEBUG0("FSMLoop: exit failure in fsmloop!");
	return EXIT_FAILURE;
      }
    }

    /*    DEBUG0("FSMLoop: Check if resedning...");*/
    if (PF_SUBTYPE(sendPack) == PST_RESEND) {
      /*
       * This subtype is set during every resend attempt.
       * Resend timeout is not decreasing during resend, but is
       * calculated from the number of packets to resend.
       */
      packNum = bufferSize - PF_PACKCNT(sendPack);
      DEBUG1("FSMLoop: Resend -> setting packNum = %d", packNum);
    }
    else if ((PF_TYPE(recvPack) == PT_DATA)) {
      /* Save DATA packet type for RTT */
      packNum = PF_PACKNUM(recvPack);
      /*      DEBUG1("FSMLoop: Storing packNum = %d", packNum);*/
    } else {
      /* Sender side or CTRL_OK/GO has been sent */
      packNum = -1;
      /*DEBUG1("FSMLoop: *** Setting packNum = %d", packNum);*/
    }

    DEBUG0("FSMLoop: Initiating new packet");
    RTTNewPack(&rTTInfo, packNum);	/* initialize RTT for a new packet */

#ifdef SERVER
    recvnBytes = -1;
#endif
  }
}

/*
 *----------------------------------------------------------------------
 * FuncTimeOut
 *
 *	Signal handler for timeouts.
 * 	Just set the flag that is looked at above when the net_recv()
 * 	returns an error (interrupted system call).
 *
 * Results:
 *	always return EXIT_SUCCESS
 *
 * Side effects:
 *	sets timeoutFlag
 *----------------------------------------------------------------------
 */

int FuncTimeOut()
{
  if (pkSent == PT_RESPONSE) {
    DEBUG0("FuncTimeOut: This server's RESPONSE was lost. Kill it!");
    exit(0);
  } else {
    timeoutFlag = TRUE;		/* set flag for function above */
    return EXIT_SUCCESS;
  }
}

/*
 *----------------------------------------------------------------------
 * int FSMInvalid
 *
 *      Invalid state transition.  Something is wrong.
 *
 * Results:
 *	return EXIT_FAILURE
 *
 * Side effects:
 *	exits FSMLoop
 *----------------------------------------------------------------------
 */

int FSMInvalid()
{
  DEBUG2("Protocol botch: pkSent = %d, pkRecv = %s",
	   pkSent, FSMName(pkRecv));
  return EXIT_FAILURE;
}

/*
 *----------------------------------------------------------------------
 * int FSMName
 *
 *      Converts packet numer type to a string name.
 *
 * Results:
 *	return string name of packet type.
 *
 * Side effects:
 *	-
 *----------------------------------------------------------------------
 */

char *FSMName(num)
     int num;
{
  static char packName[20];
  switch (num) {
  case 0:strcpy(packName,"open"); break;
  case 1:strcpy(packName,"response"); break;
  case 2:strcpy(packName,"keepalive"); break;
  case 3:strcpy(packName,"quit"); break;
  case 4:strcpy(packName,"quitack"); break;
  case 5:strcpy(packName,"abort"); break;
  case 6:strcpy(packName,"data"); break;
  case 7:strcpy(packName,"ldata"); break;
  case 8:strcpy(packName,"null_ack"); break;
  case 9:strcpy(packName,"control"); break;
  case 10:strcpy(packName,"refused"); break;
  case 11:strcpy(packName,"done"); break;
  default:strcpy(packName,"Illegal"); break;
  }
  return packName;
}






