/* 
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:	rtt.h
 * Class:	RTT
 * Created:	0.1, 970506, Per-Erik Forssn
 * Revised:	1.1, 980219, Torbjrn Frostemark
 * 
 * Description: Timer routines for round-trip timing of datagrams.
 * Remarks:	-
 *
 * Private Metods:	-
 *
 * Private Attributes:	-
 *----------------------------------------------------------------------
 */
#ifndef _RTT
#define _RTT
#include "etftp.h"

/*
 * Structure to contain everything needed for RTT timing.
 * One of these required per socket being timed.
 * The caller allocates this structure, then passes its address to
 * all the rtt_XXX() functions.
 */

typedef struct RTTStruct {
  /*float	rtt;*/		/* most recent round-trip time (RTT), seconds */
  long	rtt;
  short	rtt_nrexmt;	/* #times retransmitted: 0, 1, 2, ... */

  int wait;		/* retransmitt timer */
  int alarmTimeout;	/* time in seconds before timeout */
  
} RTTStruct;

void RTTInit(RTTStruct *ptr);
void RTTNewPack(RTTStruct *ptr, int packNum);
int RTTUpdate(RTTStruct *ptr, int packNum);
int RTTTimeout(RTTStruct *ptr);
void RTTMsleep(int uSecs);

#define	RTT_RXTMIN      2	/* min retransmit timeout value, seconds */
#define	RTT_RXTMAX    120	/* max retransmit timeout value, seconds */
#define	RTT_DEATH      10	/* max number of times to retransmit */

#endif /* _RTT */
