/* 
 * Copyright (C) 1997, 1998 Defence Research Establishment, Dept. of Comm.
 * Systems, FOA 72, Sweden.
 * Currently maintained by Christian Jnsson FOA 72 <chj@lin.foa.se>.
 *
 * This file is part of ETFTP version 1.1.3.
 *
 *   ETFTP is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   ETFTP is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with ETFTP; see the file GPL.  If not, write to the
 *   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *   Boston, MA 02111-1307, USA.
 *
 * Code:
 *
 *----------------------------------------------------------------------
 * Filenamne:	sendrecv.c
 * Class:	SendRecv
 * Created:	0.1, 970423, Henrik Brodin
 * Revised:	1.1, 980219, Torbjrn Frostemark
 * Revised:	1.1b, 980723, Torbjrn Frostemark
 * 
 * Description:	This class contains methods for recieving and sending
 *		packets.
 * Remarks:	-
 *
 * Private Attributes:   -
 *
 * Private Methods:      MakeHeader(char type)
 *----------------------------------------------------------------------
 */

#include "sendrecv.h"
#include "net.h"
#include "tron.h"
#include "file.h"
#include "etftptime.h"
#include "clu.h"
#include "rtt.h"
extern int      sys_nerr;       /* # of error message strings in sys table */
#ifndef LINUX_ETFTP
extern char     *sys_errlist[]; /* the system error message table */
#endif /* LINUX_ETFTP */
char 	errString[256];
int 	controlOK;		/* Flags for which packet to send. */
int 	controlRESEND;
int 	controlGO;
int 	seqControl;		/* How many ctrl-packets have we sent? */
TRONList *resendList;
int 	sentBytes;		/* # of sent bytes */
int	lastBufferFlag;		/* Is it time to send LDATA or LDATAL? */
int	highByteRecv;		/* Set to size of buffer when last packet
				 * arrived */
#ifdef CLIENT
int	oldBlockSize;		/* Used for comparing old and new block
				 * size values */
#endif /* CLIENT */

/*
 *----------------------------------------------------------------------
 * MakeHeader
 *
 *	Must be called AFTER SendLen is set.
 *
 * Results:
 *	Fills in the first twelve bytes of sendPack.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void MakeHeader(char pkType)
{
  unsigned int pos = 2;		/* Start with version... */
  unsigned char version = 0;
  unsigned int checkSum;
  int i;

  sendPack[pos++] = version;
  sendPack[pos++] = pkType;
  stshort(sendLen,&sendPack[pos]);
  pos += 2;
  stshort(localPort,&sendPack[pos]);
  pos += 2;
#ifdef CLIENT
  stshort(sesRemotePort,&sendPack[pos]);
#endif /* CLIENT */
#ifdef SERVER
  stshort(remotePort,&sendPack[pos]);
#endif /* SERVER */
  pos += 2;
  sendPack[pos++] = NULL_BYTE;	/* Zero blanking */
  sendPack[pos] = NULL_BYTE;	
  pos = 0;			/* Time for checksum! */
  for (i=2,checkSum=0; i < sendLen; i+=2) { /* We're counting in words,
					     *hence +=2. */
    checkSum = checkSum ^ ldshort(&sendPack[i]);
  }
  stshort(~checkSum,&sendPack[pos]);
}

/*
 *----------------------------------------------------------------------
 * SendCONTROL
 *
 *	This works by finding out which control packets the program
 *	wants to send and then build them.
 *
 * Results:
 *	Sends all CONTROL packets.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendCONTROL()
{
  int pos = 12;
  int i;

  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  DEBUG0("sendrecv: SendCONTROL");  

  if (controlOK) {
    DEBUG0("SendCONTROL: Sending CONTROL_OK");

    sendPack[pos++] = PST_OK;
    sendPack[pos++] = NULL_BYTE;

    DEBUG1("SendCONTROL: sequenceNum = %d", sequenceNum);
    stshort(sequenceNum++,&sendPack[pos]);
    pos += 2;

    /*DEBUG1("SendCONTROL(OK): Storing bufferNum = %ld", bufferNum);*/
    stlong(bufferNum,&sendPack[pos]);
    pos += 4;

    stshort(burstSize,&sendPack[pos]);
    pos += 2;

    stshort(burstRate,&sendPack[pos]);
    pos += 2;

    tightTimer = TimeGet() - tightTimer + 1;
    DEBUG1("SendCONTROL: Tighttimer value is: %d", tightTimer);
    stshort(tightTimer,&sendPack[pos]);
    pos += 2;

    stshort(blockSize,&sendPack[pos]);
    pos += 2;    
  }

  if (controlGO) {
    DEBUG0("SendCONTROL: Sending CONTROL_GO");
    sendPack[pos++] = PST_GO;
    sendPack[pos++] = NULL_BYTE;

    DEBUG1("SendCONTROL: sequenceNum = %d", sequenceNum);
    stshort(sequenceNum++,&sendPack[pos]);
    pos += 2;

    /*DEBUG1("SendCONTROL(GO): Storing bufferNum = %ld", bufferNum);*/
    stlong(bufferNum,&sendPack[pos]);
    pos += 4;

    /*DEBUG2("SendCONTROL(GO): Calling TRONReset. bufferSize = %d, totNBytes = %ld",
      bufferSize, totNBytes);*/
    TRONReset(bufferSize, totNBytes);	/* Tell TRON that a new buffer is coming.*/

    highByteRecv = 0;
  }
    
  if (controlRESEND) {
    DEBUG0("SendCONTROL: Sending CONTROL_RESEND");

    sendPack[pos++] = PST_RESEND;
    sendPack[pos++] = NULL_BYTE;

    DEBUG1("SendCONTROL: sequenceNum = %d", sequenceNum);
    stshort(sequenceNum++, &sendPack[pos]);
    pos += 2;

    /*DEBUG1("SendCONTROL(RESEND): Storing bufferNum = %ld", bufferNum);*/
    stlong(bufferNum, &sendPack[pos]);
    pos += 4;

    stshort(resendList->size, &sendPack[pos]);
    pos += 2;
    sendPack[pos++] = NULL_BYTE;
    sendPack[pos++] = NULL_BYTE;

    /*DEBUG0("SendCONTROL(RESEND): Following packets should be resent:");*/
    for ( i = 0 ; i < resendList->size ; i++) {
      stshort(resendList->list[i], &sendPack[pos]);
      pos += 2;
      /*printf("%d ", resendList->list[i]);
	fflush(stdout);*/
    }
    /*printf("\n");*/ /* New line */
    if (pos%4 > 0) {
      sendPack[pos++] = NULL_BYTE;
      sendPack[pos++] = NULL_BYTE;
    }
  }

  sendLen = pos;
  MakeHeader(PT_CONTROL);

  DEBUG1("SendCONTROL: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_CONTROL;
  controlOK=FALSE;
  controlGO=FALSE;
  controlRESEND=FALSE;

  /* The tighttimer is the time between control_GO and CONTROL_OK */
  tightTimer = TimeGet(); /* Set tight timer */
}

#ifdef CLIENT
/*
 *----------------------------------------------------------------------
 * RecvRESPONSE
 *	
 *	
 *
 * Results:
 *	
 *
 * Side effects:
 *	Initializes some variables.
 *----------------------------------------------------------------------
 */

int RecvRESPONSE()
{
  DEBUG0("sendrecv: Received RESPONSE");

  if (PF_CONNECT_ID(recvPack) != connectID) { 
    DEBUG0("RecvRESPONSE: RESPONSE from wrong ID.");
#ifdef SERVER
    FileClose(localFP);
#endif /* SERVER */
    return EXIT_FAILURE;
  }
  rtTimer = TimeGet() - rtTimer + 1;	/* Stop round trip timer */
  DEBUG1("RecvRESPONSE: Round trip timer = %ld seconds", rtTimer);

  nextBlkNum = 0;
  bufferNum = 0;
  sentBytes = 0;
  lastBufferFlag = FALSE;

  /* We take the params from the response from the server. */
  bufferSize = PF_BUFSZ(recvPack);
  blockSize = PF_DATAPACKSZ(recvPack);
  burstSize = PF_BURSTSZ(recvPack);
  burstRate = PF_BURSTRATE(recvPack);
  totNBytes = PF_TRXSZ(recvPack);
  TRONStart(burstSize, burstRate, blockSize); /* Tell TRON */
  TRONReset(bufferSize, totNBytes);
  
  /* The server wants it, we give it to him. */
  if (PF_OPENMODE(recvPack) == MODE_PUT) {
    if (hashFlag) {
      printf("Hash value: %d bytes/block\nReceiving: ", blockSize);
      fflush(stdout);
    }
    controlGO = TRUE;
    DEBUG0("RecvRESPONSE: Send CONTROL_GO");
    SendCONTROL();
  } 
  else if (hashFlag) {
    printf("Hash value: %d bytes/block\nSending: ", blockSize);
    fflush(stdout);
  }    
  return EXIT_SUCCESS;
}

/*
 *----------------------------------------------------------------------
 * RecvREFUSED
 *	Prints the string that the server sent.
 *
 * Results:
 *	
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */
int RecvREFUSED()

{
  DEBUG0("sendrecv: Received REFUSED");
  /* Print the error string from server */
  DEBUG1("RecvREFUSED: Message from server: '%s'\n", PF_REASON(recvPack));
#ifdef SERVER
  FileClose(localFP);
#endif /* SERVER */
  totNBytes = 0;		/* Since nothing has been transferred. */
  return EXIT_FAILURE;
}

/*
 *----------------------------------------------------------------------
 * SendOPEN
 *	Sends an OPEN packet to the server.
 *
 *
 * Results:
 *	EXIT_SUCCESS or EXIT_FAILURE.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int SendOPEN(fName,fSize,mode)
     char *fName;
     int fSize;
     int mode;			/* Put or get */
{
  int pos;			/* Position in packet */
  unsigned char MBZ = 0;	/* Field in OPEN */
  float tmpBurstRate;
  
  if ( traceFlag ) {
    CLUStart();			/* To initialize log. */
  }

  DEBUG2("sendrecv: Sending OPEN for %s, mode = %d", fName, mode); 

  free(sendPack);
  sendPack = calloc(MAXBUFF, 1);	/* To clear sendPack */
    
  seqControl = 0;		/* Highest seq control number received. */
  pos = 12;			/* Start of "Connection unique ID" */

  stlong(++connectID, &sendPack[pos]); /* ID ? */
  pos += 4;			/* It was a longword */

  stlong(bufferSize, &sendPack[pos]);
  DEBUG1("SendOPEN: bufferSize = %d", bufferSize);
  pos += 4;

  stlong(fSize,&sendPack[pos]); /* Size of file */
  pos += 4;

  stshort(blockSize,&sendPack[pos]);
  DEBUG1("SendOPEN: blockSize = %d", blockSize);
  pos += 2;

  stshort(burstSize,&sendPack[pos]);
  DEBUG1("SendOPEN: burstSize = %d", burstSize);
  pos += 2;

  if ((tmpBurstRate = (blockSize+72)*burstSize*8/baudRate) < 0.3) {
    burstRate = 0;
  } else {
    burstRate = (unsigned short)tmpBurstRate + 1;
  }
  DEBUG1("SendOPEN: burstRate = %d", burstRate);

  stshort(burstRate,&sendPack[pos]);
  pos += 2;

  stshort(DEATHTIME,&sendPack[pos]);
  pos += 2;

  /* Sets Reserved field (MBZ) */
  sendPack[pos++]=MBZ;
  if (mode == MODE_GET) {	/* Since we have tested for the action */
    /* DEBUG0("This is the recieving end."); */
    MBZ = MBZ | 1;
  }		
  MBZ = MBZ | 2;		/* Always checksum all */
  if (modeType == MODE_BINARY) {
    MBZ = MBZ | 4;
  }
  if (renegotiate == 1) {
    MBZ = MBZ | 8;
  }
  sendPack[pos++]=MBZ;		/* End reserved */

  stshort(multiBuff,&sendPack[pos]);	/* Maximum no. outstanding buffers */
  pos += 2;

  stshort(radioDelay,&sendPack[pos]);
  pos += 4;			/* Two unused bytes */

  strcpy(sendPack+pos, fName);
  pos += strlen(fName) + 1;

  sendLen = ((pos / 4) * 4) + 4; /* Longword alignment */
  while (pos <= sendLen) 
    sendPack[pos++] = NULL_BYTE; /* Make packet null-filled at end.*/
  MakeHeader(PT_OPEN);		/* Calculate checksum and stuff */

  NetSend(sendPack, sendLen);   /* Packet away... */

  rtTimer = TimeGet();		/* Start round trip timer */
  DEBUG0("SendOPEN: Round trip timer started...");

  pkSent = PT_OPEN;
  return EXIT_SUCCESS;		/* For historical reasons */
}
#endif /* CLIENT */

/*
 *----------------------------------------------------------------------
 * SendRESPONSE
 *
 *	
 *
 * Results:
 *	Send an positive response to the client.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */
#ifdef SERVER

void SendRESPONSE(mode, fName)
     int mode;
     char *fName;
{
  int pos;			/* Position in packet */
  unsigned char MBZ = 0;	/* Field in RESPONSE */

  DEBUG0("sendrecv: Sending RESPONSE");  

  free(sendPack);  
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  pos = 12;			/* Start of "Connection unique ID" */

  stlong(connectID,&sendPack[pos]); /* ID  */
  pos += 4;			/* It was a longword */

  stlong(bufferSize,&sendPack[pos]);
  pos += 4;

  stlong(totNBytes,&sendPack[pos]); /* Size of file */
  pos += 4;

  stshort(blockSize,&sendPack[pos]);
  pos += 2;

  stshort(burstSize,&sendPack[pos]);
  pos += 2;

  stshort(burstRate,&sendPack[pos]);
  pos += 2;

  stshort(DEATHTIME,&sendPack[pos]);
  pos += 2;

  /* Sets Reserved field (MBZ) */
  sendPack[pos++]=MBZ;

  if (mode == MODE_PUT) {
    MBZ = MBZ | 1;
    transferMode = MODE_PUT;	/* rtt stops after first DATA packet received */
  } else {
    transferMode = MODE_GET;	/* rtt stops after CONTROL(GO) */
  }

  MBZ = MBZ | 2;		/* Always checksum all */
  if (modeType == MODE_BINARY) {
    MBZ = MBZ | 4;
  }

  if (renegotiate == 1) {
    MBZ = MBZ | 8;
  }
  sendPack[pos++]=MBZ;		/* End reserved */
  stshort(1,&sendPack[pos]);	/* Maximum no. outstanding buffers. 
				 1 because of no multibuf.*/
  pos += 2;

  stshort(radioDelay,&sendPack[pos]);
  pos += 4;			/* Two unused bytes */

  strcpy(sendPack+pos, fName);
  pos += strlen(fName) + 1;

  sendLen = ((pos / 4) * 4) + 4; /* Longword alignment */
  while (pos <= sendLen) {
    sendPack[pos++] = NULL_BYTE; /* Make packet null-filled at end.*/
  }
  MakeHeader(PT_RESPONSE);		/* Calculate checksum and stuff */

  DEBUG1("SendRESPONSE: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);			/* sleep before channel turnaround */ 

  NetSend(sendPack, sendLen);		/* Packet away... */
  pkSent = PT_RESPONSE;
  nextBlkNum = 0;
  bufferNum = 0;

  rtTimer = TimeGet();		/* Start round trip timer */
  DEBUG0("SendRESPONSE: Round trip timer started...");
}
#endif /* SERVER */

/*
 *----------------------------------------------------------------------
 * SendQUITACK
 *
 *	
 *
 * Results:
 *	Acknowledges the QUIT-packet sent earlier.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendQUITACK()
{
  DEBUG0("sendrecv: Sending QUITACK");
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  sendLen = 12;
  MakeHeader(PT_QUITACK);

  DEBUG1("SendQUITACK: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_QUITACK;
}

/*
 *----------------------------------------------------------------------
 * SendDONE
 *
 *	
 *
 * Results:
 *	Sends tha packet that tells the world the transfer is over.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendDONE()
{
  DEBUG0("sendrecv: Sending DONE");
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  sendLen = 12;
  MakeHeader(PT_DONE);

  DEBUG1("SendDONE: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_DONE;
}

/*
 *----------------------------------------------------------------------
 * SendQUIT
 *
 *	
 *
 * Results:
 *	We want to shutdown the transfer.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendQUIT()
{
  short pos = 12;	/* Start on string */

  DEBUG0("sendrecv: Sending QUIT");
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  strcpy(sendPack+pos, errString); /* Reason for QUIT */
  pos += strlen(errString) + 1;

  sendLen = ((pos / 4) * 4) + 4; /* Longword alignment */
  while (pos <= sendLen) {
    sendPack[pos++] = NULL_BYTE; /* Make packet null-filled at end.*/
  }
  
  MakeHeader(PT_QUIT);		/* Calculate checksum and stuff */

  DEBUG1("SendQUIT: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_QUIT;

#ifdef CLIENT
  if (hashFlag) {
    /* print new line as end of '#' */
    putchar('\n');
    fflush(stdout);
  }
#endif /* CLIENT */
}

/*
 *----------------------------------------------------------------------
 * SendABORT
 *
 *	
 *
 * Results:
 *	Send an abort of the transfer.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendABORT()
{
  short pos = 12;	/* Start on string */
  DEBUG0("sendrecv: Sending ABORT");
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  strcpy(sendPack+pos, errString); /* Reason for ABORT */
  pos += strlen(errString) + 1;
  
  sendLen = ((pos / 4) * 4) + 4; /* Longword alignment */
  while (pos <= sendLen) {
    sendPack[pos++] = NULL_BYTE; /* Make packet null-filled at end.*/
  }
  
  MakeHeader(PT_ABORT);	/* Calculate checksum and stuff */

  DEBUG1("SendABORT: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_ABORT;
}

/*
 *----------------------------------------------------------------------
 * SendREFUSED
 *
 *	
 *
 * Results:
 *	If we for some reason want to refuse the transfer.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */
#ifdef SERVER

void SendREFUSED()
{
  short pos = 12;	/* Start on string */
  DEBUG0("SendREFUSED: Sending REFUSED");
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */
  
  strcpy(sendPack+pos, errString); /* Reason for REFUSED */
  pos += strlen(errString) + 1;
  
  sendLen = ((pos / 4) * 4) + 4; /* Longword alignment */
  while (pos <= sendLen) {
    sendPack[pos++] = NULL_BYTE; /* Make packet null-filled at end.*/
  }
  
  MakeHeader(PT_REFUSED);	/* Calculate checksum and stuff */

  DEBUG1("SendREFUSED: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack, sendLen);
  pkSent = PT_REFUSED;  
}
#endif /* SERVER */

/*
 *----------------------------------------------------------------------
 * SendPacket
 *
 *	
 *
 * Results:
 *	Transmits a given packet in the current buffer.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendPacket(number,type)
     int number;
     int type;
{
  unsigned int 	pos;
  int j, startByte, checkSum, stopByte;		/* Counters */
  char MBZ = 0;

  DEBUG2("SendPacket: SendPacket. number = %d, type = %d", number, type);

  free(sendPack);
  sendPack = calloc(MAXBUFF, 1);	/* To clear sendPack */  

  pos = 12;			/* Start on buffernumber */
  stlong(bufferNum, &sendPack[pos]);
  pos +=4;
  stshort(seqControl, &sendPack[pos]);
  pos +=2;
  stshort(number, &sendPack[pos]);
  pos +=4;
  sendPack[pos++] = NULL_BYTE; 
  if (lastBufferFlag) {
    MBZ = MBZ | 1;		/* DATAL or LDATAL packet */
  }
  sendPack[pos++] = MBZ;
  
  startByte = blockSize*number; /* Time to copy DATA */
  stopByte = startByte + blockSize;
  
  if (blockSize+sentBytes > totNBytes) { /* blockSize was stopSize, wrong */
    /* last packet in file */
    stopByte = totNBytes - sentBytes;
    DEBUG2("SendPacket: Calculating last stopByte = %d. sentBytes = %d",
	   stopByte, sentBytes);
  }

  for (j = startByte ; j <= stopByte ; j++) {
    sendPack[pos++] = sendBuff[j];
  }
  sendLen = pos - 1;
  for ( j = 0 ; j < sendLen%4 ; j++) {
    sendPack[sendLen+j] = NULL_BYTE;
  }
  pos = 20;			/* This is a step backwards */
  checkSum = sendPack[24];	/* Data area checksum */
  for ( j = 26 ; j <= sendLen; j += 2) { /* We're counting in words, hence +=2. */
    checkSum = checkSum ^ ldshort(&sendPack[j]);
  }
  stshort(~checkSum, &sendPack[pos]);

  if ((number == (nextBlkNum - 1)) || (type == 7)) { /* Last block in buffer */
    MakeHeader(PT_LDATA);
    pkSent = PT_LDATA;
    DEBUG0("SendPacket: Sending LDATA");
    DEBUG4("SendPacket: bufferNum = %ld. Buffer byte %d - %d. DATA packet nr %d",
	   bufferNum, startByte, stopByte, PF_PACKNUM(sendPack));
  } else {
    MakeHeader(PT_DATA);
    pkSent = PT_DATA;
    DEBUG4("SendPacket: bufferNum = %ld. Buffer bytes %d - %d. DATA packet nr %d",
	   bufferNum, startByte, stopByte, PF_PACKNUM(sendPack));
  }
  if((sendLen%4) != 0) {
    sendLen += 4 - (sendLen%4);
  }
  NetSend(sendPack, sendLen);
}

/*
 *----------------------------------------------------------------------
 * SendBuffer
 *
 *	
 *
 * Results:
 *	Chops the buffer into packets and then transmits them.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendBuffer()
{
  unsigned int 	pos;
  int i, j, startByte, checkSum, stopByte, burstNum = 0;	/* Counters */
  int firstDataNum;
  unsigned int lDATASleep;                             /* Timer fixed by ChJ */

  char MBZ;

  DEBUG0("***********************************************");
  DEBUG0("sendrecv: Starting buffer transfer (SendBuffer).");
  while (nextBlkNum < bufferSize) {
    /* This means that we will continue for the whole buffer */
    /* RTTTimeGet(&tV); */
    free(sendPack);
    sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

    DEBUG2("SendBuffer: Radio delay -> waiting %d seconds before burst %d...", radioDelay, burstNum++);
    sleep(radioDelay);

    firstDataNum = nextBlkNum;
    DEBUG1("SendBuffer: firstDataNum = %d", firstDataNum);

    for (i = 0 ; i < burstSize ; i++) {
      MBZ = 0;
      pos = 12;			/* Start on buffernumber */

#ifdef CLIENT
      if (hashFlag) {
	/* print '#' for each sent DATA packet in client */
	putchar('#');
	fflush(stdout);
      }
#endif /* CLIENT */

      stlong(bufferNum, &sendPack[pos]);
      pos +=4;
      stshort(seqControl, &sendPack[pos]);	
      pos +=2;
      stshort(nextBlkNum, &sendPack[pos]);
      pos +=4;
      sendPack[pos++] = NULL_BYTE; 
      if (lastBufferFlag) {
	MBZ = MBZ | 1;		/* DATAL Packet */
      }
      sendPack[pos++] = MBZ;
      
      startByte = blockSize * nextBlkNum; /* Time to copy DATA */
      stopByte = startByte + blockSize;

      if (stopByte+sentBytes > totNBytes) {
	stopByte = totNBytes - sentBytes;
      }
      for (j = startByte ; j <= stopByte ; j++) {
	sendPack[pos++] = sendBuff[j];	
      }

      sendLen = pos-1;
   
      for(j = 0; j < sendLen%4; j++) {
	  sendPack[sendLen + j]=NULL_BYTE;
	}
      pos = 20;
      checkSum = sendPack[24];	/* Data area checksum */
      for(j=26; j <= sendLen; j+=2) { /* We're counting in words, hence +=2. */
	checkSum = checkSum ^ ldshort(&sendPack[j]);
      }
      stshort(~checkSum,&sendPack[pos]);
      nextBlkNum++;

      DEBUG3("SendBuffer: stopByte=%d, sentBytes=%d, totNBytes=%ld",
	     stopByte, sentBytes, totNBytes);

      if ((nextBlkNum == bufferSize) || (stopByte+sentBytes == totNBytes)) {
	/* End of buffer -> Send LDATA */
	MakeHeader(PT_LDATA);	/* Fill in the first 12 bytes  */
	pkSent = PT_LDATA;

	if ((sendLen%4) != 0) {
	  sendLen += 4 - (sendLen%4);
	}

	DEBUG0("SendBuffer: Sending LDATA");
        DEBUG4("SendBuffer: bufferNum = %ld. Buffer byte %d - %d. DATA packet nr %d",
	       bufferNum, startByte, stopByte, PF_PACKNUM(sendPack));

	NetSend(sendPack,sendLen);

	lDATASleep = (unsigned int)                    /* Timer fixed by ChJ */
	  ( burstRate * ((float)(nextBlkNum - firstDataNum) / burstSize));
	
	DEBUG1("SendBuffer: LDATA-sleeping %d s...", lDATASleep);
	sleep(lDATASleep);
	return;

      } else {
	/* Middle of buffer -> Send DATA */
	MakeHeader(PT_DATA);	/* Fill in the first 12 bytes */
	if (TRUE) {	/* OBS! Debug! */
	  pkSent = PT_DATA;
	  if(sendLen%4!=0) {
	    sendLen+=4-(sendLen%4);
	  }
 	  DEBUG4("SendBuffer: bufferNum = %ld. Buffer bytes %d - %d. DATA packet nr %d", bufferNum, startByte,
		 stopByte, PF_PACKNUM(sendPack));
	  NetSend(sendPack,sendLen);
	}
      }
    } /* for loop */

    DEBUG1("SendBuffer: Sleeping %d s...", burstRate);
    DEBUG2("SendBuffer: burstRate = %d, burstSize = %d", burstRate, burstSize);
    sleep(burstRate);
  } 
}

/*
 *----------------------------------------------------------------------
 * SendNULL_ACK
 *	Send an acknowledge of a CONTROL_OK.
 *	
 *
 * Results:
 *	
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

void SendNULL_ACK()
{
  unsigned int pos = 12;

  DEBUG0("sendrecv: Sending NULL_ACK");  
  free(sendPack);
  sendPack = calloc(MAXBUFF,1);	/* To clear sendPack */

  stshort(seqControl, &sendPack[pos]);
  pos +=2;
  stshort(burstSize, &sendPack[pos]);
  pos +=2;
  stshort(burstRate, &sendPack[pos]);
  pos +=2;
  stshort(blockSize, &sendPack[pos]);
  sendLen = pos + 2;
  MakeHeader(PT_NULL_ACK);

  DEBUG3("SendNULL_ACK: BurstSize %d, BurstRate %d, BlockSize %d",burstSize,burstRate,blockSize);
  /* DEBUG1("SendNULL_ACK: New Msleeptime: %d",burstRate/burstSize); */

  DEBUG1("SendNULL_ACK: Sleeping %d s before channel turnaround.", radioDelay);
  sleep(radioDelay);		/* sleep before channel turnaround */

  NetSend(sendPack,sendLen);
  pkSent = PT_NULL_ACK;
}

/*
 *----------------------------------------------------------------------
 * RecvCONTROL
 *
 *
 *
 * Results:
 *	
 *
 * Side effects:
 *	Sets some variables.
 *----------------------------------------------------------------------
 */

int RecvCONTROL()
{
  int 	pos = 12;
  int 	readNBytes, i, oldpos, size, send;

  DEBUG0("sendrecv: RecvCONTROL");
  while (pos < PF_LENGTH(recvPack)-2) { /* If last was NULL-padded */
    oldpos = pos;

    if (recvPack[oldpos] == PST_GO) {
      DEBUG0("RecvCONTROL: Received CONTROL_GO");

#ifdef SERVER
      if ((transferMode == MODE_GET) && rttSet) {
	/* If server is transmitter rtt is stopped here */
	rtTimer = TimeGet() - rtTimer + 1;
	DEBUG1("RecvCONTROL(GO): Round trip timer = %ld", rtTimer);
	rttSet = FALSE;
      }
#endif /* SERVER */
      
      if (seqControl > PF_SEQNUM(recvPack)) {
	/* this CONTROL_GO packet is old -> skip it */
	DEBUG2("RecvCONTROL(GO): Current control seq. number = %d, received = %d -> skip!",
	       seqControl, PF_SEQNUM(recvPack));
	return EXIT_SUCCESS;
      }

      if ((pkSent == PT_LDATA) && !(nextBlkNum == 0)) {
	/* Check if Control OK or if duplicated Control Go */
	DEBUG0("RecvCONTROL(GO): Special timeout... double Control_GO");
	return -1;
      }

      if (PF_SEQNUM(recvPack) > seqControl) {
	seqControl = PF_SEQNUM(recvPack);
	DEBUG1("RecvCONTROL(GO): Increasing control sequence number = %d", seqControl);
      }
      
      if (sentBytes+bufferSize*blockSize > totNBytes) {
	DEBUG0("RecvCONTORL(GO): Setting lastBufferFlag!");
	lastBufferFlag = TRUE;
      }
      pos += 8;
      
      DEBUG3("RecvCONTROL(GO): totNBytes = %ld, sentBytes = %d -> %ld bytes to go",
	     totNBytes, sentBytes, totNBytes - sentBytes);
      
      if ((totNBytes - sentBytes) < (bufferSize * blockSize)) {
	/* read last buffer */
	readNBytes = totNBytes - sentBytes;
      } else {
	/* read a regular buffer */
	readNBytes = bufferSize * blockSize;
      }	
      
      FileRead(localFP, sendBuff, readNBytes, modeType);
      DEBUG1("RecvCONTROL(GO): Reading %d bytes", readNBytes);
      DEBUG1("RecvCONTROL(GO): At file position: %d", sentBytes);

      DEBUG1("RecvCONTROL(GO): Sleeping %d s before channel turnaround.", radioDelay);
      sleep(radioDelay);		/* sleep before channel turnaround */

      SendBuffer();
    }
    
    if (recvPack[oldpos] == PST_OK) {
      DEBUG0("RecvCONTROL: Received CONTROL_OK");
      
      if ((bufferNum >= PF_BUFNUM_REQ(recvPack))) {
	/* this CONTROL_OK packet has allready been processed */
	DEBUG2("RecvCONTROL(OK): Current buffer number = %ld, received = %ld -> skip!",
	       bufferNum, PF_BUFNUM_REQ(recvPack));
	return EXIT_SUCCESS;
      }

      if (seqControl > PF_SEQNUM(recvPack)) {
	/* this CONTROL_OK packet is old -> skip it */
	DEBUG2("RecvCONTROL(OK): Current control seq. number = %d, received = %d -> skip!",
	       seqControl, PF_SEQNUM(recvPack));
	return EXIT_SUCCESS;
      }

      if (PF_SEQNUM(recvPack) > seqControl) {
	seqControl = PF_SEQNUM(recvPack);
	DEBUG1("RecvCONTROL(OK): Increasing control sequence number = %d", seqControl);
      }

      bufferNum = PF_BUFNUM_REQ(recvPack);
      DEBUG1("RecvCONTROL(OK): New bufferNum = %ld", bufferNum);

      if (lastBufferFlag) {
	/* whole file has received */
	sentBytes = totNBytes;
      } else {
	/* still receiving file */
	sentBytes += bufferSize*blockSize;
      }
      
      /*
       * Check if new sending parms. If so -> send NULL_ACK.
       */
      send = FALSE;
      if (burstSize != PF_BURSTSZ_REQ(recvPack)) {
	burstSize = PF_BURSTSZ_REQ(recvPack);
	DEBUG1("RecvCONTROL(OK): *** New burstSize = %d", burstSize);
	send = TRUE; /* send NULL_ACK */
      }
      if (burstRate != PF_BURSTRATE_REQ(recvPack)) {
	burstRate = PF_BURSTRATE_REQ(recvPack);
	DEBUG1("RecvCONTROL(OK): *** New burstRate = %d", burstRate);
	send = TRUE; /* send NULL_ACK */
      }
	if (blockSize != PF_PACKSZ_REQ(recvPack)) {
	blockSize = PF_PACKSZ_REQ(recvPack);
	DEBUG1("RecvCONTROL(OK): *** New blockSize = %d", blockSize);
	send = TRUE; /* send NULL_ACK */
      }
      if (send) {
	DEBUG0("RecvCONTROL(OK): Parms have changed. Send NULL_ACK.");
	SendNULL_ACK();		/* If the params have changed, send NULL_ACK */
      }

      looseTimer = PF_CTRLTIMER(recvPack);
      nextBlkNum = 0;		/* Remember, we're numbering inside buffers */
      pos += 16;
      if (lastBufferFlag) {
/*#ifdef SERVER*/
	FileClose(localFP);	/* Here we close before DONE is received. */
/*#endif*/ /* SERVER */
	DEBUG0("RecvCONTROL(OK): Last buffer received. Send QUIT");
	SendQUIT();
	return EXIT_SUCCESS;
      }
    }

    if (recvPack[oldpos] == PST_RESEND) {
      DEBUG0("RecvCONTROL: Received CONTROL_RESEND");

      if (seqControl > PF_SEQNUM(recvPack)) {
	/* this CONTROL_RESEND packet is old -> skip it */
	DEBUG2("RecvCONTROL(RESEND): Current control seq. number = %d, received = %d -> skip!",
	       sequenceNum, PF_SEQNUM(recvPack));
	return EXIT_SUCCESS;
      }

      if (PF_SEQNUM(recvPack) > seqControl) {
	seqControl = PF_SEQNUM(recvPack);
	DEBUG1("RecvCONTROL(RESEND): Increasing control sequence number = %d", seqControl);
      }

      pos += 12;
      size = PF_PACKCNT(recvPack); /* number of packets to resend */

      /*      if (size && verboseFlag) {*/
	/* DEBUG1("RecvCONTROL(RESEND): Following %d packet(s) will be resent:", size);*/
      /*	for (i = 0 ; i < size ; i++) { 
		printf("%d ", ldshort(&recvPack[i*2+pos])); 
		}
		printf("\n");
		}*/

      DEBUG1("RecvCONTROL(RESEND): Sleeping %d s before channel turnaround.", radioDelay);
      sleep(radioDelay);		/* sleep before channel turnaround */

      for (i = size - 1 ; i >= burstSize ; i -= burstSize) { 
	int j;
	for (j = 0 ; j < burstSize ; j++) {
	  /* Send burst */
	  DEBUG0("RecvCONTROL(RESEND): Resending DATA packet");
	  SendPacket(ldshort(&recvPack[pos]), 6);
	  pos += 2;
	}
	/* Sleep between bursts */
	sleep(burstRate);
	sleep(radioDelay);	/* user induced sleep */
      }
      for (; i > 0 ; i--) {
	/* Send last burst */
	DEBUG0("RecvCONTROL(RESEND): Resending DATA packet");
	SendPacket(ldshort(&recvPack[pos]), 6);
	pos += 2;
      }
      /* Send LDATA */
      DEBUG0("RecvCONTROL(RESEND): Resending LDATA packet");
      SendPacket(ldshort(&recvPack[pos]), 7);
      pos += 2;
    }
  }
  return EXIT_SUCCESS;
}


/*
 *----------------------------------------------------------------------
 * RecvOPEN
 *
 *	
 *
 * Results:
 *	Processes the OPEN packet sent by client.
 *
 * Side effects:
 *	Creates or open an local file (localFP is set).
 *----------------------------------------------------------------------
 */

#ifdef SERVER

int RecvOPEN()
{
  char dirName[MAXFILENAME],fileName[MAXFILENAME];
  struct stat statBuff;
  int mode;
  unsigned short recvBlockSize, recvBurstRate, recvBurstSize;

  DEBUG0("sendrecv: Received OPEN");
  
  if (connected) {
    strcpy(errString,"Already connected with an other client.");
    SendREFUSED(); 
    return EXIT_FAILURE;
  }
  if (recvPack[ldshort(&recvPack[4]) - 1] != '\0') {
    strcpy(errString,"Invalid filename.");
    SendREFUSED(); /* Filename not null-terminated */
    return EXIT_FAILURE;
  }
  strcpy(fileName, PF_FILENAME(recvPack));
  if (fileName[0] != '/') {
    strcpy(errString,"Filename must begin with '/'.");
    SendREFUSED();
    return EXIT_FAILURE;
  }
  connected = TRUE;
  sentBytes = 0;
  lastBufferFlag = FALSE;

  remotePort = PF_LOCALPORT(recvPack);
  connectID = PF_CONNECT_ID(recvPack);
  bufferSize = PF_BUFSZ(recvPack);
  totNBytes = PF_TRXSZ(recvPack); /* File size in bytes */

  recvBlockSize = PF_DATAPACKSZ(recvPack);
  recvBurstRate = PF_BURSTRATE(recvPack);
  recvBurstSize = PF_BURSTSZ(recvPack);

  /* Server has absolute power when it comes to upper inital block size */
  
  if (blockSize >= recvBlockSize) {
    /* Requested parms are OK */
    burstRate = recvBurstRate;
    blockSize = recvBlockSize;
  } else {
    float tmpBurstRate;

    DEBUG2("RecvOPEN: Requested block size (%d) too high. Setting new block size (%d)", recvBlockSize, blockSize);
    
    tmpBurstRate = recvBurstRate * (blockSize + 72) / (recvBlockSize + 72);
    
    if (tmpBurstRate < 0.3) {
      burstRate = 0;
    } else {
      burstRate = (unsigned short)tmpBurstRate + 1;
    }
  }
  
  if (burstSize != recvBurstSize) {
    /* New burstRate does not have to be calculated since the new burstSize is
     * already in recvBurstRate, hence now in burstRate */
    DEBUG2("RecvOPEN: Old burst size %d. New burst size %d", burstSize, recvBurstSize);
  }
  burstSize = recvBurstSize;
  
  DEBUG1("RecvOPEN: New burstRate = %d.", burstRate);
  
  if (PF_RENEGOTIATE(recvPack)) {
    DEBUG0("RecvOPEN: Enable renegotiate");
    renegotiate = 1;
  } else {
    DEBUG0("RecvOPEN: Disable renegotiate");
    renegotiate = FALSE;
  }

  if (PF_TRMODE(recvPack)) {
    DEBUG0("BINARY MODE");
    modeType = MODE_BINARY;
  } else {
    DEBUG0("ASCII MODE");
    modeType = MODE_ASCII;
  }

  TRONStart(burstSize,burstRate,blockSize); /* Tell TRON */
  TRONReset(bufferSize, totNBytes);

  /* Now it's time to do something about that file transfer. */
  if (PF_OPENMODE(recvPack) == MODE_GET) {	/* We're sending... */
    DEBUG1("RecvOPEN: Received request to read file '%s'",fileName);
    mode = MODE_GET;
    if (stat(fileName, &statBuff) < 0) {
      strcpy(errString,SysErrStr());
      DEBUG1("%s",errString);
      DEBUG0("RecvOPEN: Send REFUSED");
      SendREFUSED();
      return EXIT_FAILURE;
    }
    else if ((statBuff.st_mode & (S_IREAD >> 6)) == 0) {
      strcpy(errString,"File doesn't allow world read permission");
      DEBUG1("%s",errString);
      SendREFUSED();
      DEBUG0("RecvOPEN: Send REFUSED");
      return EXIT_FAILURE;
    }
    else if ((localFP = FileOpen(fileName)) == NULL) {
      strcpy(errString,ERR_NOFILE);
      DEBUG1("%s",errString);
      DEBUG0("RecvOPEN: Send REFUSED");
      SendREFUSED();
      return EXIT_FAILURE;
    }
    totNBytes = FileCount(fileName, modeType);	/* Since the server knows the filesize */
    ldshort(&recvPack[34]);	/* Max no. outstanding buffers... */
    radioDelay = PF_RADIODELAY(recvPack);
    DEBUG0("RecvOPEN: Send RESPONSE");
    SendRESPONSE(mode, fileName);

  } else {			/* We're receiving... */
    DEBUG1("RecvOPEN: Received request to write file '%s'.",fileName);
    mode = MODE_PUT;
    strcpy(dirName, fileName);
    *(strrchr(dirName, '/') + 1) = '\0'; /* Strip filename */
  
    if (stat(dirName, &statBuff) < 0) {
      strcpy(errString, SysErrStr());
      DEBUG1("%s", errString);
      DEBUG0("RecvOPEN: Send REFUSED");
      SendREFUSED();
      return EXIT_FAILURE;
    }
    else if ((statBuff.st_mode & (S_IWRITE >> 6)) == 0) {
      strcpy(errString, "Directory doesn't allow world write permission");
      DEBUG1("%s", errString);
      DEBUG0("RecvOPEN: Send REFUSED");
      SendREFUSED();
      return EXIT_FAILURE;
    }
    else if ((localFP = FileCreate(fileName)) == NULL) {
      strcpy(errString, ERR_NOFILE);
      DEBUG1("%s", errString);
      DEBUG0("RecvOPEN: Send REFUSED");
      SendREFUSED();
      return EXIT_FAILURE;
    }
    ldshort(&recvPack[34]);	/* Max no. outstanding buffers... */
    radioDelay = PF_RADIODELAY(recvPack);
    DEBUG0("RecvOPEN: Send RESPONSE");
    SendRESPONSE(mode, fileName);
    controlGO = TRUE;
    DEBUG0("RecvOPEN: Send CONTROL");
    SendCONTROL();
  }

  pkRecv = PT_OPEN;
  DEBUG0("RecvOPEN: Request accepted.");
  return EXIT_SUCCESS;
}
#endif /* SERVER */

/*
 *----------------------------------------------------------------------
 * RecvQUITACK
 *	Receives QUITACK.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int RecvQUITACK()
{
  DEBUG0("sendrecv: Received QUITACK");
  pkRecv = PT_QUITACK;
#ifdef SERVER
  FileClose(localFP);
#endif /* SERVER */
  return EXIT_FAILURE;
}

/*
 *----------------------------------------------------------------------
 * RecvDONE
 *	Takes the DONE packet and returns EXIT_FAILURE to tell
 *      FSMLoop to exit.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int RecvDONE()
{
  DEBUG0("sendrecv: Received DONE");
  pkRecv = PT_DONE;
  return EXIT_FAILURE;
}

/*
 *----------------------------------------------------------------------
 * RecvQUIT
 *	Prints the errorstring from the other side.
 *
 * Results:
 *	
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int RecvQUIT()
{
  DEBUG0("sendrecv: Received QUIT");
  pkRecv = PT_QUIT;
  if (strlen(PF_REASON(recvPack))) {
    DEBUG1("RecvQUIT: Quit-message from counterpart: '%s'", PF_REASON(recvPack));
  }
  if (lastBufferFlag) {
    DEBUG0("RecvQUIT: Last buffer received -> send DONE.");
    SendDONE();			/* We must tell the other side that we're done */
  } else {
#ifdef SERVER
    FileClose(localFP);
#endif /* SERVER */
    DEBUG0("RecvQUIT: Send QUITACK!");
    SendQUITACK();
  }

#ifdef CLIENT
  if (hashFlag) {
    /* print new line as end of '#' */
    putchar('\n');
    fflush(stdout);
  }
#endif /* CLIENT */

  return EXIT_FAILURE;
  /*return EXIT_SUCCESS;*/
}

/*
 *----------------------------------------------------------------------
 * RecvABORT
 *	Prints error string received.
 *	
 *
 * Results:
 *	
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int RecvABORT()
{
  DEBUG0("sendrecv: Received ABORT");
  pkRecv = PT_ABORT;
  if (PF_REASON(recvPack) != NULL) {
    DEBUG1("RecvABORT: Abort-message from counterpart: '%s'", PF_REASON(recvPack));
  }
#ifdef SERVER
  FileClose(localFP);
#endif /* SERVER */
  return EXIT_FAILURE;
}

/*
 *----------------------------------------------------------------------
 * RecvDATA
 *
 *	
 *
 * Results:
 *	Processes incoming DATA packet.
 *
 * Side effects:
 *	Buffer is filled.
 *----------------------------------------------------------------------
 */

int RecvDATA()
{
  int	startByte, j, pos;
  struct RecvStat	statistic;
  pos = 24;

  DEBUG0("sendrecv: RecvDATA");
  if (PF_LASTFLAG(recvPack)) {
    DEBUG0("RecvDATA: Setting last buffer flag!");
    lastBufferFlag = TRUE;
  }

#ifdef CLIENT
  if (hashFlag) {
    /* print '#' for each received DATA packet */
    putchar('#');
    fflush(stdout);
    }
  bytesReceived += PF_LENGTH(recvPack) - 24;	/* data - header */
#endif /* CLIENT */

#ifdef SERVER
  if ((transferMode == MODE_PUT) && rttSet) {
    /* If server is sender rtt is stopped here */
    /* rtTimer is corrected for the bigger DATA packet measured on */
    rtTimer = TimeGet() - rtTimer - burstRate/burstSize + 1;
    DEBUG1("RecvDATA: Round trip timer = %ld", rtTimer);
    rttSet = FALSE;
  }
#endif /* SERVER */

  startByte = blockSize*PF_PACKNUM(recvPack); /* Time to copy DATA */
  for (j = startByte ; j < startByte+blockSize ; j++) {
    recvBuff[j] = recvPack[pos++];
  }

  seqControl = PF_TOPNUM(recvPack);
  /*DEBUG1("RecvDATA: Control packet %d is acknowledged", seqControl);*/

  DEBUG4("RecvDATA: bufferNum = %ld. Buffer bytes %d - %d. DATA packet nr %d", PF_BUFNUM(recvPack), 
	 startByte, startByte + (PF_LENGTH(recvPack) - 24), PF_PACKNUM(recvPack));
  statistic.aveRate = TRONUpdate(PF_PACKNUM(recvPack), PF_LENGTH(recvPack) - 24);
 
 if ( traceFlag ) {
    CLUPut(&statistic);		/* if traceflag print statistics */
  }
  return EXIT_SUCCESS;
}

/*
 *----------------------------------------------------------------------
 * RecvLDATA
 *
 *	
 *
 * Results:
 *	Receives the Last DATA packet in the current buffer.
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

int RecvLDATA()
{
  int	startByte, j, pos;
  struct RecvStat	statistic;

  pos = 24;

  DEBUG0("sendrecv: RecvLDATA");

#ifdef CLIENT
  if (hashFlag) {
    /* print '#' for each received DATA packet in client */
    putchar('#');
    fflush(stdout);
    }
  bytesReceived += PF_LENGTH(recvPack) - 24;	/* data - header */
#endif /* CLIENT */

#ifdef SERVER
      if ((transferMode == MODE_PUT) && rttSet) {
	/* If server is sender and only LDATA packet
	 * is sent rtt is stopped here */
	/* rtTimer is corrected for the bigger DATA packet measured on */
	rtTimer = TimeGet() - rtTimer - burstRate/burstSize + 1;
	DEBUG1("RecvLDATA: Round trip timer = %ld", rtTimer);
	rttSet = FALSE;
      }
#endif /* SERVER */

  if (PF_LASTFLAG(recvPack)) {
    DEBUG0("RecvLDATA: Setting last buffer flag!");
    lastBufferFlag = TRUE;
  }

  seqControl = PF_TOPNUM(recvPack);
  DEBUG1("RecvLDATA: Control packet %d is acknowledged", seqControl);

  DEBUG2("RecvLDATA: bufferNum = %ld, packet buffer = %ld", bufferNum, PF_BUFNUM(recvPack));
  if (bufferNum > PF_BUFNUM(recvPack)) {
    /* server has resent LDATA -> special case */
    DEBUG0("RecvLDATA: Server has resent LDATA -> skip it!");
    DEBUG0("RecvLDATA: Received a buffer -> send CONTROL_OK");    
    controlOK = TRUE;
    if (!lastBufferFlag) {
      DEBUG0("RecvLDATA: Not last buffer -> send CONTROL_GO");
      controlGO = TRUE;
    }
    SendCONTROL();
    return EXIT_SUCCESS;
  }

  startByte = blockSize*PF_PACKNUM(recvPack); /* Time to copy DATA */
  for (j = startByte ; j < startByte+PF_LENGTH(recvPack)-24 ; j++) {
    recvBuff[j] = recvPack[pos++];
  }
  if (j > highByteRecv) {
    highByteRecv = j;
  }
  DEBUG3("RecvLDATA: Received LDATA. Buffer bytes %d - %d. DATA packet nr %d", startByte,
	 startByte + (PF_LENGTH(recvPack) - 24), PF_PACKNUM(recvPack));

#ifdef CLIENT
  DEBUG2("RecvLDATA: Bytes received so far = %ld -> %ld to go...", bytesReceived,
	 totNBytes - bytesReceived);
  oldBlockSize = blockSize;	/* store current block size */
#endif /* CLIENT */

  pkRecv = PT_LDATA;
  DEBUG0("RecvLDATA: Update TRON");
  statistic.aveRate = TRONUpdate(PF_PACKNUM(recvPack), PF_LENGTH(recvPack) - 24);
  if ( traceFlag ) {
    CLUPut(&statistic);
  }
  resendList = TRONResend();
  DEBUG1("RecvLDATA: No of packets to resend = %d", resendList->size);

  if (resendList->size == 0) {
    /* Complete buffer transfered */
    nextBlkNum = 0;
    bufferNum++;
    DEBUG1("RecvLDATA: Increasing bufferNum. New buffer = %ld", bufferNum);

    if (!lastBufferFlag) {
      /* if not last buffer in transfer -> send CONTROL_GO */
      controlGO = TRUE;
    }

    if (renegotiate) {
      DEBUG0("RecvLDATA: Calling TRONDecide");
      TRONDecide(&burstSize, &burstRate, &blockSize);
    }

    FileWrite(localFP, recvBuff, highByteRecv, modeType);
    DEBUG1("RecvLDATA: Wrote %d bytes ", highByteRecv);

    controlOK = TRUE;	/* always send CONTROL_OK */

    if (lastBufferFlag) {
      FileClose(localFP);	/* Note that the file is closed */
				/* before NULL-ACK packet is received. */
    }
  } else {
    /* Some packets are missing in current buffer -> send CONTROL_RESEND */
    DEBUG1("RecvLDATA: Missing packets in buffer %ld -> Send CONTROL_RESEND", bufferNum);
    controlRESEND = TRUE;
  }

  SendCONTROL();
  return EXIT_SUCCESS;
}

/*
 *----------------------------------------------------------------------
 * RecvNULL_ACK
 *	We have gotten ourselves a NULL_ACK.
 *	
 *
 * Results:
 *	
 *
 * Side effects:
 *	Sets some variables.
 *----------------------------------------------------------------------
 */

int RecvNULL_ACK()
{
  DEBUG0("sendrecv: Received NULL_ACK");
  burstSize = PF_BURSTSZ_ACK(recvPack);
  burstRate = PF_BURSTRATE_ACK(recvPack);
  blockSize = PF_PACKSZ_ACK(recvPack);
    
#ifdef CLIENT
  if ((blockSize != oldBlockSize) && hashFlag) {
    /* If new block size and hash on -> show new hash value */
    printf("\nNew hash value: %d bytes/block\n", blockSize);
    fflush(stdout);
  }
#endif /* CLIENT */
    
  DEBUG3("RecvNULL_ACK: BurstSize %d, BurstRate %d, BlockSize %d", burstSize, burstRate, blockSize);
  pkRecv = PT_NULL_ACK;
  return EXIT_SUCCESS;
}

/*
 *----------------------------------------------------------------------
 * SysErrStr
 *
 * 	Return a string containing some additional operating-system
 * 	dependent information.
 * 	Note that different versions of UNIX assign different meanings
 * 	to the same value of "errno" (compare errno's starting with 35
 * 	between System V and BSD, for example).  This means that if an error
 * 	condition is being sent to another UNIX system, we must interpret
 * 	the errno value on the system that generated the error, and not
 * 	just send the decimal value of errno to the other system.
 *	
 *	This is a ripoff from tftp.
 *
 * Results:
 *	Returns an string containig system errors
 *
 * Side effects:
 *	None.
 *----------------------------------------------------------------------
 */

char *SysErrStr()
{
  static char	msgstr[200];
  
  if (errno != 0) {
    if (errno > 0 && errno < sys_nerr) {
      sprintf(msgstr, "%s", sys_errlist[errno]);
    } else {
      sprintf(msgstr, "errno %d", errno);
    }
  } else {
    msgstr[0] = '\0';
  }

#ifdef SYS5
  if (t_errno != 0) {
    char	tmsgstr[100];
    
    if (t_errno > 0 && t_errno < sys_nerr) {
      sprintf(tmsgstr, " %s", t_errlist[t_errno]);
    } else {
      sprintf(tmsgstr, ", t_errno %d", t_errno);
    }    
    strcat(msgstr, tmsgstr);
  }
#endif

  return msgstr;
}

int RecvTimeout()
{
  DEBUG0("sendrecv: Timed out, no LDATA");
  resendList = TRONResend();
  if (resendList->size == 0) {
    DEBUG0("Something is fishy in RecvTimeout, should never be called. Contact your nearest ETFTP dealer.");
  } else {
    controlRESEND = TRUE;
  }
  SendCONTROL();
  return EXIT_SUCCESS;
}

/* E N D   O F   L I N E */
